/*
 * apple2fs driver for apple2 ucsd disk file system access
 * copyright (c)2003 GPL Oskar Schirmer (oskar@scara.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/blkdev.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/stat.h>
#include <linux/time.h>
#include "apple2fs.h"

void apple2fs_put_inode(struct inode *inode)
{
	kfree(inode->u.generic_ip);
}

void apple2fs_read_inode (struct inode *inode)
{
	struct super_block *sb;
	struct apple2fs_direntry dire;
	struct apple2fs_file_info pd;
	int filenr;
	unsigned int filetype;
	int i;

	if (inode->i_ino < APPLE2FS_ROOT_INO) {
		printk ("APPLE2FS: bad inode number %ld\n", inode->i_ino);
		make_bad_inode(inode);
		return;
	}

	sb = inode->i_sb;
	filenr = APPLE2FS_INO_FNDX(inode->i_ino);
	if (filenr > APPLE2FS_SBU_INFO(sb,nfiles)) {
		/* this entry is unused */
		make_bad_inode(inode);
		return;
	}

	if (apple2fs_getrootdirentry(sb, filenr, &dire) < 0) {
		make_bad_inode(inode);
		return;
	}

	filetype = dire.filetype;
	if ((inode->i_ino >= APPLE2FS_INO_FILE(1))
	 && ((inode->i_ino & APPLE2FS_INO_MASK) == APPLE2FS_TEXT_INO)
	 && (filetype != APPLE2FS_FTYPE_TEXT)) {
		/* this text inode is not a text file */
		make_bad_inode(inode);
		return;
	}

	inode->u.generic_ip = (void *)kmalloc(
			sizeof(struct apple2fs_inode_info), GFP_KERNEL);
	if (inode->u.generic_ip == NULL) {
		printk("APPLE2FS: kmalloc failed for inode\n");
		make_bad_inode(inode);
		return;
	}

	switch (filetype) {
	case APPLE2FS_FTYPE_DIR:
		inode->i_mode = S_IRUGO | S_IXUGO | S_IFDIR;
		break;
	case APPLE2FS_FTYPE_CODE:
		inode->i_mode = S_IRUGO | S_IXUGO | S_IFREG;
		break;
	default:
		inode->i_mode = S_IRUGO | S_IFREG;
		break;
	}
	inode->i_uid = 0;
	inode->i_gid = 0;
	if (inode->i_ino == APPLE2FS_ROOT_INO) {
		inode->i_nlink = 2 + APPLE2FS_INO_MASK;
	} else if (inode->i_ino < APPLE2FS_INO_FILE(1)) {
		inode->i_nlink = 2;
	} else {
		inode->i_nlink = 1;
	}
	APPLE2FS_IPU_INFO(inode,firstblock) = dire.firstblk;
	i = dire.followblk - dire.firstblk;
	inode->i_blocks = i;
	i *= APPLE2FS_BLOCKSIZE;
	if (filetype == APPLE2FS_FTYPE_DIR) {
		i -= APPLE2FS_ROOTDIR * APPLE2FS_BLOCKSIZE;
	} else if ((inode->i_ino >= APPLE2FS_INO_FILE(1))
		&& ((inode->i_ino & APPLE2FS_INO_MASK) == APPLE2FS_TEXT_INO)) {
		pd.phys_pos = APPLE2FS_TEXTPHYSBEGIN;
		pd.leftover_dle = 0;
		pd.accum_null = 0;
		i = apple2fs_scantextfile(sb, &pd, dire.firstblk,
			-APPLE2FS_BLOCKSIZE * inode->i_blocks, NULL);
	} else {
		i += dire.u.file.usedbytesinlastblk
			- APPLE2FS_BLOCKSIZE;
	}
	inode->i_size = i; /* FIXME: wrong for text inode */
	if (filetype == APPLE2FS_FTYPE_DIR) {
		i = dire.u.dir.lastbootdate;
	} else {
		i = dire.u.file.moddate;
	}
	inode->i_atime = inode->i_ctime = inode->i_mtime =
		APPLE2FS_DATE_YEAR(i) < 70 ? 0 :
		mktime(1900 + APPLE2FS_DATE_YEAR(i),
			APPLE2FS_DATE_MONTH(i), APPLE2FS_DATE_DAY(i), 0, 0, 0);
	inode->i_version = ++event;
	inode->i_generation = 0;

	if (filetype == APPLE2FS_FTYPE_DIR) {
		inode->i_op = &apple2fs_dir_inode_operations;
		inode->i_fop = &apple2fs_dir_operations;
	} else {
		inode->i_fop = &apple2fs_file_operations;
	}
}
